/*
    Copyright (C) 2010 wpitchoune@gmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include <sensors/sensors.h>
#include <sensors/error.h>

#include "config.h"
#include "hdd.h"
#include "psensor.h"
#include "graph.h"
#include "nvidia.h"
#include "ui.h"
#include "ui_sensorlist.h"
#include "ui_color.h"
#include "lmsensor.h"
#include "ui_pref.h"

#ifdef HAVE_APP_INDICATOR
#include "ui_appindicator.h"
#endif

#ifdef HAVE_LIBNOTIFY
#include "ui_notify.h"
#endif

#include "compat.h"

#ifndef PSENSOR_VERSION
#define PSENSOR_VERSION "0.0.0"
#endif

#ifndef PSENSOR_DIR
#define PSENSOR_DIR "/usr/share/psensor"
#endif

#define PSENSOR_ICON_FILENAME "psensor-32x32.png"

static void
cb_preferences(gpointer data, guint callback_action, GtkWidget * item)
{
	struct ui_psensor *ui_psensor = (struct ui_psensor *)data;

	ui_pref_dialog_run(ui_psensor);
}

static GtkItemFactoryEntry menu_items[] = {
	{"/Preferences",
	 NULL, cb_preferences, 0, "<Item>"},

	{"/sep1",
	 NULL, NULL, 0, "<Separator>"},

	{"/Quit",
	 "", gtk_main_quit, 0, "<StockItem>", GTK_STOCK_QUIT},
};

static gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);

void update_information(struct psensor **sensors)
{
	while (1) {
		gdk_threads_enter();

		if (!sensors)
			return;

#ifdef NVIDIA_SUPPORT

		nvidia_psensor_list_update(sensors);

#endif

		lmsensor_psensor_list_update(sensors);

		if (psensor_list_contains_type(sensors, SENSOR_TYPE_HDD_TEMP))
			hdd_psensor_list_update(sensors);

		gdk_threads_leave();

		sleep(1);
	}
}

gboolean refresh_thread(gpointer data)
{
	struct ui_psensor *ui_psensor = (struct ui_psensor *)data;

	gdk_threads_enter();

	update_graph(ui_psensor->sensors,
		     ui_psensor->w_graph, ui_psensor->config);

	ui_sensorlist_update(ui_psensor->ui_sensorlist);

#ifdef HAVE_APP_INDICATOR
	ui_appindicator_update(ui_psensor);
#endif

	gdk_threads_leave();

	return TRUE;
}

gboolean
on_expose_event(GtkWidget * widget, GdkEventExpose * event, gpointer data)
{
	struct ui_psensor *ui_psensor = (struct ui_psensor *)data;

	update_graph(ui_psensor->sensors,
		     ui_psensor->w_graph, ui_psensor->config);

	return FALSE;
}

void on_destroy(GtkWidget * widget, gpointer data)
{
	gtk_main_quit();
}

void cb_alarm_raised(struct psensor *sensor, void *data)
{
#ifdef HAVE_LIBNOTIFY
	if (sensor->enabled) {
		struct ui_psensor *ui = data;

		ui_notify(sensor, ui);
	}
#endif
}

GdkPixbuf *create_pixbuf(const char *filename)
{
	GdkPixbuf *pixbuf;
	GError *err = NULL;

	pixbuf = gdk_pixbuf_new_from_file(filename, &err);

	if (pixbuf) {
		return pixbuf;
	} else {
		fprintf(stderr, "ERROR: while loading icon: %s\n",
			err->message);
		g_error_free(err);
		return NULL;
	}
}

char *get_psensor_icon_path()
{
	int pathlength
	    = strlen(PSENSOR_DIR) + 1 + strlen(PSENSOR_ICON_FILENAME) + 1;
	char *path = malloc(pathlength);

	strcpy(path, PSENSOR_DIR);

	if (PSENSOR_DIR[strlen(PSENSOR_DIR) - 1] != '/')
		strcat(path, "/");

	strcat(path, PSENSOR_ICON_FILENAME);

	return path;
}

GdkPixbuf *create_psensor_icon()
{
	char *path = get_psensor_icon_path();

	GdkPixbuf *pixbuf = create_pixbuf(path);

	free(path);

	return pixbuf;
}

void associate_colors(struct psensor **sensors)
{
	/* number of uniq colors */
#define COLORS_COUNT 8

	unsigned int colors[COLORS_COUNT][3] = {
		{0x0000, 0x0000, 0x0000},	/* black */
		{0xffff, 0x0000, 0x0000},	/* red */
		{0x0000, 0.0000, 0xffff},	/* blue */
		{0x0000, 0xffff, 0x0000},	/* green */

		{0x7fff, 0x7fff, 0x7fff},	/* grey */
		{0x7fff, 0x0000, 0x0000},	/* dark red */
		{0x0000, 0x0000, 0x7fff},	/* dark blue */
		{0x0000, 0x7fff, 0x0000}	/* dark green */
	};

	struct psensor **sensor_cur = sensors;
	int i = 0;
	while (*sensor_cur) {
		struct color default_color;
		color_set(&default_color,
			  colors[i % COLORS_COUNT][0],
			  colors[i % COLORS_COUNT][1],
			  colors[i % COLORS_COUNT][2]);

		(*sensor_cur)->color
		    = config_get_sensor_color((*sensor_cur)->id,
					      &default_color);

		sensor_cur++;
		i++;
	}
}

void associate_cb_alarm_raised(struct psensor **sensors, struct ui_psensor *ui)
{
	struct psensor **sensor_cur = sensors;
	while (*sensor_cur) {
		struct psensor *s = *sensor_cur;

		s->cb_alarm_raised = cb_alarm_raised;
		s->cb_alarm_raised_data = ui;

		if (is_temp_type(s->type)) {
			s->alarm_limit
			    = config_get_sensor_alarm_limit(s->id, 60);
			s->alarm_enabled
			    = config_get_sensor_alarm_enabled(s->id);
		} else {
			s->alarm_limit = 0;
			s->alarm_enabled = 0;
		}

		sensor_cur++;
	}
}

void associate_preferences(struct psensor **sensors)
{
	struct psensor **sensor_cur = sensors;
	while (*sensor_cur) {
		char *n;
		struct psensor *s = *sensor_cur;

		s->enabled = config_is_sensor_enabled(s->id);

		n = config_get_sensor_name(s->id);

		if (n)
			s->name = n;

		sensor_cur++;
	}
}

GtkWidget *ui_get_popupmenu(gpointer data)
{
	GtkItemFactory *item_factory;
	GtkWidget *menu;

	item_factory = gtk_item_factory_new(GTK_TYPE_MENU, "<main>", NULL);
	gtk_item_factory_create_items(item_factory,
				      nmenu_items, menu_items, data);
	menu = gtk_item_factory_get_widget(item_factory, "<main>");

	return menu;
}

int on_graph_clicked(GtkWidget * widget, GdkEventButton * event, gpointer data)
{
	GtkWidget *menu;

	if (event->type != GDK_BUTTON_PRESS)
		return FALSE;

	menu = ui_get_popupmenu(data);

	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
		       event->button, event->time);

	return TRUE;
}

GtkWidget *ui_window_create(struct ui_psensor * ui)
{
	GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GdkScreen *screen;
	GdkColormap *colormap;
	GdkPixbuf *icon;

	gtk_window_set_default_size(GTK_WINDOW(window), 800, 200);

	gtk_window_set_title(GTK_WINDOW(window),
			     "Psensor - Temperature Monitor");

	screen = gtk_widget_get_screen(window);

	if (ui->config->alpha_channel_enabled
	    && gdk_screen_is_composited(screen)
	    && (colormap = gdk_screen_get_rgba_colormap(screen)))
		gtk_widget_set_colormap(window, colormap);
	else
		ui->config->alpha_channel_enabled = 0;

	icon = create_psensor_icon();
	if (icon)
		gtk_window_set_icon(GTK_WINDOW(window), icon);

	g_signal_connect(window, "destroy", G_CALLBACK(on_destroy), NULL);

	gtk_window_set_decorated(GTK_WINDOW(window),
				 ui->config->window_decoration_enabled);

	gtk_window_set_keep_below(GTK_WINDOW(window),
				  ui->config->window_keep_below_enabled);

	return window;
}

GtkWidget *create_graph_widget(struct ui_psensor * ui)
{
	GtkWidget *w_graph;

	w_graph = gtk_drawing_area_new();

	g_signal_connect(G_OBJECT(w_graph),
			 "expose-event", G_CALLBACK(on_expose_event), ui);

	gtk_widget_add_events(w_graph, GDK_BUTTON_PRESS_MASK);
	gtk_signal_connect(GTK_OBJECT(w_graph),
			   "button_press_event",
			   (GCallback) on_graph_clicked, ui);

	return w_graph;
}

void ui_main_box_create(struct ui_psensor *ui)
{
	struct config *cfg;
	GtkWidget *w_sensorlist;

	cfg = ui->config;

	if (ui->main_box) {
		ui_sensorlist_create_widget(ui->ui_sensorlist);

		gtk_container_remove(GTK_CONTAINER(ui->main_window),
				     ui->main_box);

		ui->w_graph = create_graph_widget(ui);
		ui->w_sensorlist = ui->ui_sensorlist->widget;
	}

	if (cfg->sensorlist_position == SENSORLIST_POSITION_RIGHT
	    || cfg->sensorlist_position == SENSORLIST_POSITION_LEFT)
		ui->main_box = gtk_hpaned_new();
	else
		ui->main_box = gtk_vpaned_new();

	w_sensorlist = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w_sensorlist),
				       GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(w_sensorlist),
			  ui->ui_sensorlist->widget);

	gtk_container_add(GTK_CONTAINER(ui->main_window), ui->main_box);

	if (cfg->sensorlist_position == SENSORLIST_POSITION_RIGHT
	    || cfg->sensorlist_position == SENSORLIST_POSITION_BOTTOM) {
		gtk_paned_pack1(GTK_PANED(ui->main_box),
				GTK_WIDGET(ui->w_graph), TRUE, TRUE);
		gtk_paned_pack2(GTK_PANED(ui->main_box),
				w_sensorlist, FALSE, TRUE);
	} else {
		gtk_paned_pack1(GTK_PANED(ui->main_box),
				w_sensorlist, FALSE, TRUE);
		gtk_paned_pack2(GTK_PANED(ui->main_box),
				GTK_WIDGET(ui->w_graph), TRUE, TRUE);
	}

	gtk_widget_show_all(ui->main_box);
}

int main(int argc, char **argv)
{
	struct ui_psensor ui;
	GError *error;
	GThread *thread;
	int err;

	printf("Psensor - Temperature Monitor - v%s by wpitchoune@gmail.com\n",
	       PSENSOR_VERSION);
	printf("Released under GPLv2 license terms, read copyright file\n");
	printf("http://wpitchoune.net/psensor\n");

	g_thread_init(NULL);
	gdk_threads_init();
	gdk_threads_enter();

	gtk_init(&argc, &argv);

#ifdef HAVE_LIBNOTIFY
	ui.notification_last_time = NULL;
#endif

	config_init();

	ui.config = config_load();

	err = lmsensor_init();
	if (!err) {
		fprintf(stderr, "ERROR: lmsensor init failure: %s\n",
			sensors_strerror(err));
		return 1;
	}

	ui.sensors = get_all_sensors();
	associate_preferences(ui.sensors);
	associate_colors(ui.sensors);
	associate_cb_alarm_raised(ui.sensors, &ui);

	/* main window */
	ui.main_window = ui_window_create(&ui);
	ui.main_box = NULL;

	/* drawing box */
	ui.w_graph = create_graph_widget(&ui);

	/* sensor list */

	ui.ui_sensorlist = ui_sensorlist_create(ui.sensors);

	ui_main_box_create(&ui);

	gtk_widget_show_all(ui.main_window);

	thread = g_thread_create((GThreadFunc) update_information,
				 ui.sensors, TRUE, &error);

	if (!thread)
		g_error_free(error);

	g_timeout_add(1000, refresh_thread, &ui);

#ifdef HAVE_APP_INDICATOR
	ui_appindicator_init(&ui);
#endif

	/* main loop */
	gtk_main();

	sensors_cleanup();

	psensor_list_free(ui.sensors);

	gdk_threads_leave();

	return 0;
}
